# Copyright 2011 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
from httplib import HTTPResponse as BaseHTTPResponse
from httplib import HTTPConnection as BaseHTTPConnection

class FileSocket(object):
    """ I adapt a pair of file objects to look like a socket """

    def __init__(self, rfile, wfile):
        self.rfile = rfile
        self.wfile = wfile

    def sendall(self, data):
        self.wfile.write(data)
        self.wfile.flush()

    def makefile(self, mode, flags):
        if mode.startswith("r"):
            return self.rfile
        raise NotImplementedError

    def close(self):
        pass


class HTTPResponse(BaseHTTPResponse):

    def close(self):
        pass


class HTTPConnection(BaseHTTPConnection):

    response_class = HTTPResponse

    def __init__(self, rfile=sys.stdin, wfile=sys.stdout):
        self.rfile = rfile
        self.wfile = wfile

        BaseHTTPConnection.__init__(self, "stdio")

    def connect(self):
        self.sock = FileSocket(self.rfile, self.wfile)




from yaybu.harness import FakeChrootTestCase
import subprocess
import time
import os

class GitTest(FakeChrootTestCase):
    """
    Test the git checkout provider.

    To run these tests, git must be installed in the test environment.

    For the moment these tests depend upon the test environment having network
    access; this should ideally change in future.
    """
    # Assume presence of a master branch in the repos below
    UPSTREAM_REPO = "git://github.com/isotoma/isotoma.recipe.django.git"
    UPSTREAM_REPO_2 = "git://github.com/isotoma/yaybu.git"

    OTHER_UPSTREAM_REF = "version3"

    def test_clone(self):
        CLONED_REPO = "/tmp/test_clone"
        self.fixture.check_apply("""
            resources:
                - Checkout:
                    scm: git
                    name: %(clone_dir)s
                    repository: %(repo_url)s
                    branch: master
            """ % {
                "clone_dir": CLONED_REPO,
                "repo_url": self.UPSTREAM_REPO,
            }
        )

    def test_change_branch(self):
        """Test for a change in branch after an initial checkout """

        CLONED_REPO = "/tmp/test_change_branch"

        # Do the initial checkout
        self.fixture.check_apply("""
            resources:
                - Checkout:
                    scm: git
                    name: %(clone_dir)s
                    repository: %(repo_url)s
                    branch: master
            """ % {
                "clone_dir": CLONED_REPO,
                "repo_url": self.UPSTREAM_REPO,
            }
        )

        # Change to another ref
        self.fixture.check_apply("""
            resources:
                - Checkout:
                    scm: git
                    name: %(clone_dir)s
                    repository: %(repo_url)s
                    branch: %(alt_ref)s
            """ % {
                "clone_dir": CLONED_REPO,
                "repo_url": self.UPSTREAM_REPO,
                "alt_ref": self.OTHER_UPSTREAM_REF,
            }
        )

    def test_checkout_tag(self):
        """Test checking out a particular tag, when specifying the tag within
        the branch field in the Checkout mapping"""

        CLONED_REPO = "/tmp/test_checkout_tag"
        TAG = "0.1.0"

        self.fixture.check_apply("""
            resources:
                - Checkout:
                    scm: git
                    name: %(clone_dir)s
                    repository: %(repo_url)s
                    branch: %(tag)s
            """ % {
                "clone_dir": CLONED_REPO,
                "repo_url": self.UPSTREAM_REPO_2,
                "tag": TAG,
            }
        )


    def test_branch_to_tag(self):
        """Test checking out a branch, then changing to a tag, then back again"""

        CLONED_REPO = "/tmp/test_branch_to_tag"
        TAG = "0.1.0"
        BRANCH = "master"

        framework = """
            resources:
                - Checkout:
                    scm: git
                    name: %(clone_dir)s
                    repository: %(repo_url)s
                    branch: %(branch_or_tag)s
            """

        branch = framework % {
            "clone_dir": CLONED_REPO,
            "repo_url": self.UPSTREAM_REPO_2,
            "branch_or_tag": BRANCH,
        }

        tag = framework % {
            "clone_dir": CLONED_REPO,
            "repo_url": self.UPSTREAM_REPO_2,
            "branch_or_tag": TAG,
        }

        self.fixture.check_apply(branch)
        self.fixture.check_apply(tag)
        self.fixture.check_apply(branch)

    def test_change_repo(self):
        """Test for the edge-case where a different repository must be checked
        out into the same location as one that already exists."""

        CLONED_REPO = "/tmp/test_change_repo"

        self.fixture.check_apply("""
            resources:
                - Checkout:
                    scm: git
                    name: %(clone_dir)s
                    repository: %(repo_url)s
                    branch: master
            """ % {
                "clone_dir": CLONED_REPO,
                "repo_url": self.UPSTREAM_REPO,
            }
        )

        self.fixture.check_apply("""
            resources:
                - Checkout:
                    scm: git
                    name: %(clone_dir)s
                    repository: %(repo_url)s
                    branch: master
            """ % {
                "clone_dir": CLONED_REPO,
                "repo_url": self.UPSTREAM_REPO_2,
            }
        )

    def test_checkout_revision(self):
        """Check out a particular revision"""

        CLONED_REPO = "/tmp/test_checkout_revision"

        self.fixture.check_apply("""
            resources:
                - Checkout:
                    scm: git
                    name: %(clone_dir)s
                    repository: %(repo_url)s
                    revision: e24b4af3710201b011ba19752176645dcd9b0edc
            """ % {
                "clone_dir": CLONED_REPO,
                "repo_url": self.UPSTREAM_REPO,
            }
        )



# Copyright 2011 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from yaybu.core.resource import Resource
from yaybu.core.policy import (
    Policy,
    Absent,
    Present,
    XOR,
    NAND)

from yaybu.core.argument import (
    FullPath,
    String,
    Integer,
    Octal,
    File,
    Dict,
    List,
    )


# Let's take a minute to thank the greybeards for this little gem...
import os
default_umask = os.umask(0)
os.umask(default_umask)


class Execute(Resource):

    """ Execute a command. This command is not executed in a shell - if you
    want a shell, run it (for example bash -c).

    For example::

        Execute:
          name: core_packages_apt_key
          command: apt-key adv --keyserver keyserver.ubuntu.com --recv-keys ${source.key}

    A much more complex example. This shows executing a command if a checkout synchronises::

        Execute.foreach bi in ${flavour.base_images}:
          name: base-image-${bi}
          policy:
              apply:
                  when: sync
                  on: /var/local/checkouts/ci
          command: ./vmbuilder-${bi}
          cwd: /var/local/checkouts/ci
          user: root

    """

    name = String()
    """ The name of this resource. This should be unique and descriptive, and
    is used so that resources can reference each other. """

    command = String()
    """ If you wish to run a single command, then this is the command. """

    commands = List()
    """ If you wish to run multiple commands, provide a list """

    cwd = FullPath(default='/')
    """ The current working directory in which to execute the command. """

    environment = Dict()
    """

    The environment to provide to the command, for example::

        Execute:
            name: example
            command: echo $FOO
            environment:
                FOO: bar
    """

    returncode = Integer(default=0)
    """ The expected return code from the command, defaulting to 0. If the
    command does not return this return code then the resource is considered
    to be in error. """

    user = String(default="root")
    """ The user to execute the command as.
    """

    group = String(default="root")
    """ The group to execute the command as.
    """

    umask = Octal(default=default_umask)
    """ The umask to use when executing this command """

    unless = String(default="")
    """ A command to run to determine is this execute should be actioned
    """

    creates = FullPath()
    """ The full path to a file that execution of this command creates. This
    is used like a "touch test" in a Makefile. If this file exists then the
    execute command will NOT be executed. """

    touch = FullPath()
    """ The full path to a file that yaybu will touch once this command has
    completed successfully. This is used like a "touch test" in a Makefile. If
    this file exists then the execute command will NOT be executed. """

class ExecutePolicy(Policy):

    """ Execute the the command or commands provided.

    If user or group attributes are provided the command will be run using sudo."""

    resource = Execute
    name = "execute"
    default = True
    signature = (Present("name"),
        XOR(Present("command"), Present("commands")),
        XOR(Present("creates"), Present("touch")),
        )




import os, shutil

from yaybu.harness import FakeChrootTestCase
from yaybu.util import sibpath
from yaybu.core import error


class TestUser(FakeChrootTestCase):

    def test_simple_user(self):
        self.fixture.check_apply("""
            resources:
                - User:
                    name: test
            """)

    def test_disabled_login(self):
        self.fixture.check_apply("""
            resources:
                - User:
                    - name: test
                      disabled-login: True
            """)
        rv = self.fixture.apply("""
            resources:
                - User:
                    - name: test
                      disabled-login: True
            """)
        self.assertEqual(rv, 254)

    def test_user_with_home(self):
        self.fixture.check_apply("""
            resources:
                - User:
                    name: test
                    home: /home/foo
            """)

    def test_user_with_impossible_home(self):
        rv = self.fixture.apply("""
            resources:
                - User:
                    name: test
                    home: /does/not/exist
            """)
        self.assertEqual(rv, error.UserAddError.returncode)

    def test_user_with_uid(self):
        self.fixture.check_apply("""
            resources:
                - User:
                    name: test
                    uid: 1111
            """)

    def test_user_with_gid(self):
        self.fixture.check_apply("""
            resources:
                - Group:
                    name: testgroup
                    gid: 1111
                - User:
                    name: test
                    gid: 1111
            """)

    def test_user_with_fullname(self):
        self.fixture.check_apply("""
            resources:
                - User:
                    name: test
                    fullname: testy mctest
            """)

    def test_user_with_password(self):
        self.fixture.check_apply("""
            resources:
                - User:
                    name: test
                    password: password
            """)

    def test_user_with_group(self):
        self.fixture.check_apply("""
            resources:
                - User:
                    name: test
                    group: nogroup
            """)

    def test_user_with_groups(self):
        self.fixture.check_apply("""
            resources:
                - User:
                    name: test
                    groups:
                        - nogroup
            """)

    def test_user_with_groups_replace(self):
        self.fixture.check_apply("""
            resources:
                - User:
                    name: test
                    groups:
                        - nogroup
                    append: False
            """)


class TestUserRemove(FakeChrootTestCase):

    def test_remove_existing(self):
        self.failUnless(self.fixture.get_user("nobody"))

        self.fixture.check_apply("""
            resources:
                - User:
                    name: nobody
                    policy: remove
            """)

        self.failUnlessRaises(KeyError, self.fixture.get_user, "nobody")

    def test_remove_non_existing(self):
        self.failUnlessRaises(KeyError, self.fixture.get_user, "zzidontexistzz")

        rv = self.fixture.apply("""
            resources:
                - User:
                    name: zzidontexistzz
                    policy: remove
            """)

        self.failUnlessEqual(rv, 254)

        self.failUnlessRaises(KeyError, self.fixture.get_user, "zzidontexistzz")




# Copyright 2011 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import optparse
import yay
from yaybu.core import runner, remote, runcontext
import logging, atexit

def version():
    import pkg_resources

    yaybu_version = pkg_resources.get_distribution('Yaybu').version
    yay_version = pkg_resources.get_distribution('Yay').version
    return 'Yaybu %s\n' \
           'yay %s' % (yaybu_version, yay_version)

def main():
    parser = optparse.OptionParser(version=version())
    parser.add_option("-s", "--simulate", default=False, action="store_true")
    parser.add_option("-p", "--ypath", default=[], action="append")
    parser.add_option("", "--log-facility", default="2", help="the syslog local facility number to which to write the audit trail")
    parser.add_option("", "--log-level", default="info", help="the minimum log level to write to the audit trail")
    parser.add_option("-d", "--debug", default=False, action="store_true", help="switch all logging to maximum, and write out to the console")
    parser.add_option("-l", "--logfile", default=None, help="The filename to write the audit log to, instead of syslog. Note: the standard console log will still be written to the console.")
    parser.add_option("-v", "--verbose", default=2, action="count", help="Write additional informational messages to the console log. repeat for even more verbosity.")
    parser.add_option("--host", default=None, action="store", help="A host to remotely run yaybu on")
    parser.add_option("-u", "--user", default="root", action="store", help="User to attempt to run as")
    parser.add_option("--remote", default=False, action="store_true", help="Run yaybu.protocol client on stdio")
    parser.add_option("--ssh-auth-sock", default=None, action="store", help="Path to SSH Agent socket")
    parser.add_option("--expand-only", default=False, action="store_true", help="Set to parse config, expand it and exit")
    parser.add_option("--resume", default=False, action="store_true", help="Resume from saved events if terminated abnormally")
    parser.add_option("--no-resume", default=False, action="store_true", help="Clobber saved event files if present and do not resume")
    parser.add_option("--env-passthrough", default=[], action="append", help="Preserve an environment variable in any processes Yaybu spawns")
    opts, args = parser.parse_args()

    if len(args) != 1:
        parser.print_help()
        return 1

    if opts.debug:
        opts.logfile = "-"
        opts.verbose = 2

    if opts.expand_only:
        ctx = runcontext.RunContext(args[0], opts)
        cfg = ctx.get_config().get()

        if opts.verbose <= 2:
            cfg = dict(resources=cfg.get("resources", []))

        print yay.dump(cfg)
        return 0

    if opts.ssh_auth_sock:
        os.environ["SSH_AUTH_SOCK"] = opts.ssh_auth_sock

    atexit.register(logging.shutdown)

    # Probably not the best place to put this stuff...
    if os.path.exists("/etc/yaybu"):
        config = yay.load_uri("/etc/yaybu")
        opts.env_passthrough = config.get("env-passthrough", opts.env_passthrough)

    if opts.host:
        r = remote.RemoteRunner()
        r.load_system_host_keys()
        r.set_missing_host_key_policy("ask")
    else:
        r = runner.Runner()

    if not opts.remote:
        ctx = runcontext.RunContext(args[0], opts)
    else:
        ctx = runcontext.RemoteRunContext(args[0], opts)

    rv = r.run(ctx)
    return rv




# Copyright 2011 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import stat
import pwd
import grp
import logging

from yaybu import resources
from yaybu.core import provider, error

from yaybu.providers.filesystem.files import AttributeChanger

class Directory(provider.Provider):

    policies = (resources.directory.DirectoryAppliedPolicy,)

    @classmethod
    def isvalid(self, *args, **kwargs):
        return super(Directory, self).isvalid(*args, **kwargs)

    def check_path(self, directory, simulate):
        frags = directory.split("/")
        path = "/"
        for i in frags:
            path = os.path.join(path, i)
            if not os.path.exists(path):
                if self.resource.parents:
                    return
                if simulate:
                    return
                raise error.PathComponentMissing(path)
            if not os.path.isdir(path):
                raise error.PathComponentNotDirectory(path)

    def apply(self, context):
        changed = False
        self.check_path(os.path.dirname(self.resource.name), context.simulate)
        ac = AttributeChanger(context,
                              self.resource.name,
                              self.resource.owner,
                              self.resource.group,
                              self.resource.mode)
        if not os.path.exists(self.resource.name):
            command = ["/bin/mkdir"]
            if self.resource.parents:
                command.append("-p")
            command.append(self.resource.name.encode("utf-8"))
            context.shell.execute(command)
            changed = True
        ac.apply(context)
        if changed or ac.changed:
            return True
        else:
            return False

class RemoveDirectory(provider.Provider):

    policies = (resources.directory.DirectoryRemovedPolicy,)

    @classmethod
    def isvalid(self, *args, **kwargs):
        return super(RemoveDirectory, self).isvalid(*args, **kwargs)

    def apply(self, context):
        if os.path.exists(self.resource.name) and not os.path.isdir(self.resource.name):
            raise error.InvalidProviderError("%r: %s exists and is not a directory" % (self, self.resource.name))
        if os.path.exists(self.resource.name):
            context.shell.execute(["/bin/rmdir", self.resource.name])
            changed = True
        else:
            changed = False
        return changed

class RemoveDirectoryRecursive(provider.Provider):

    policies = (resources.directory.DirectoryRemovedRecursivePolicy,)

    @classmethod
    def isvalid(self, *args, **kwargs):
        return super(RemoveDirectoryRecursive, self).isvalid(*args, **kwargs)

    def apply(self, context):
        if os.path.exists(self.resource.name) and not os.path.isdir(self.resource.name):
            raise error.InvalidProviderError("%r: %s exists and is not a directory" % (self, self.resource.name))
        if os.path.exists(self.resource.name):
            context.shell.execute(["/bin/rm", "-rf", self.resource.name])
            changed = True
        else:
            changed = False
        return changed




# Copyright 2011 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os, shlex, glob

from yaybu.core import provider
from yaybu.providers.service import utils
from yaybu import resources


class _LsbServiceMixin(utils._ServiceMixin):

    @classmethod
    def isvalid(cls, policy, resource, yay):
        if not super(_LsbServiceMixin, cls).isvalid(policy, resource, yay):
            return False
        if getattr(resource, policy.name):
            return False
        if os.path.exists("/sbin/start") and os.path.exists("/etc/init/%s.conf" % resource.name):
            return False
        return os.path.exists("/etc/init.d/%s" % resource.name)

    def get_command(self, action):
        return ["/etc/init.d/%s" % self.resource.name, action]

    def _enabled_links(self):
        for x in (2, 3, 4, 5):
            yield "/etc/rc%s.d/S%02d%s" % (x, self.resource.priority, self.resource.name)
        for x in (0, 1, 6):
            yield "/etc/rc%s.d/K%02d%s" % (x, 100-self.resource.priority, self.resource.name)

    def _disabled_links(self):
        for x in (0, 1, 2, 3, 4, 5, 6):
            yield "/etc/rc%s.d/K%02d%s" % (x, 100-self.resource.priority, self.resource.name)

    def _update_links(self, context, goal):
        # We turn our "goal" symlinks into a set and use a glob to get a set of
        # all symlinks in an rc.d directory for the current service name
        # The difference between the 2 sets are the links we need to create
        # and the links we need to remove
        target = set(goal)
        current = set(glob.glob("/etc/rc*.d/[SK][0-9][0-9]%s"  % self.resource.name))

        need_deleting = current - target
        need_creating = target - current

        if not need_deleting and not need_creating:
            return False

        for ln in need_deleting:
            context.shell.execute(["rm", ln])

        for ln in need_creating:
            context.shell.execute(["ln", "-s", "/etc/init.d/%s" % self.resource.name, ln])

        return True

    def ensure_enabled(self, context):
        return self._update_links(context, self._enabled_links())

    def ensure_disabled(self, context):
        return self._update_links(context, self._disabled_links())


class Start(_LsbServiceMixin, utils._Start, provider.Provider):
    pass


class Stop(_LsbServiceMixin, utils._Stop, provider.Provider):
    pass


class Restart(_LsbServiceMixin, utils._Restart, provider.Provider):
    pass




# Copyright 2011 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from urlparse import urlparse, parse_qs

from yaybu.core.protocol.server import HttpResource
from yaybu.core import error

class FileResource(HttpResource):

    leaf = True

    def render_GET(self, yaybu, request, restpath):
        params = parse_qs(request.getargs)

        try:
            # Always read in binary mode. Opening files in text mode may cause
            # newline translations, making the actual size of the content
            # transmitted *less* than the content-length!
            f = yaybu.get_file(params["path"][0])
        except error.MissingAsset:
            request.send_error(404, "File not found")
            return None

        request.send_response(200, "OK")
        request.send_header("Content-Type", "application/octect-stream")
        request.send_header("Content-Length", str(f.len))
        #request.send_header("Last-Modified", self.date_time_string(fs.st_mtime))
        request.send_header("Content", "keep-alive")
        request.end_headers()

        request.write_fileobj(f)


class EncryptedResource(HttpResource):

    leaf = True

    def render_GET(self, yaybu, request, restpath):
        contents = yaybu.get_decrypted_file(restpath).read()

        request.send_response(200, "OK")
        request.send_header("Content-Type", "application/octect-stream")
        request.send_header("Content-Length", str(len(contents)))
        request.send_header("Content", "keep-alive")
        request.end_headers()

        request.wfile.write(contents)




# Copyright 2011 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# (This is to test that resources we define outsite of yaybu.resources can
# be picked up by whitehat)
from yaybu.core import resource
class WhiteHat(resource.Resource):
    pass

import unittest
from yaybu.core.whitehat import *


def dummy_function(sitename):
    Directory(
        name = '{sitename}',
        )


class TestWhitehat(unittest.TestCase):

    def setUp(self):
        reset_bundle()

    def at(self, idx):
        return get_bundle().values()[idx]

    def test_custom_resources(self):
        WhiteHat(name='hello')
        self.failUnlessEqual(self.at(0).name, 'hello')

    def test_simple_locals(self):
        local1 = 'hello'
        local2 = '0755'

        File(
            name = '/etc/{local1}',
            mode = '{local2}',
            )

    def test_for_each(self):
        for i in range(3):
            Service(
                name = 'zope{i}',
                )

        self.failUnlessEqual(self.at(0).name, 'zope0')
        self.failUnlessEqual(self.at(1).name, 'zope1')
        self.failUnlessEqual(self.at(2).name, 'zope2')

    def test_function(self):
        dummy_function('/www.foo.com')
        self.failUnlessEqual(self.at(0).name, '/www.foo.com')

    def test_inner_function(self):
        def dummy_function(somevar):
            Directory(
                name = '/www.foo.com-{somevar}',
                )
        dummy_function('test')

        self.failUnlessEqual(self.at(0).name, '/www.foo.com-test')




# Copyright 2011 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Resources dealing with filesystem objects other than files. """

from yaybu.core.resource import Resource
from yaybu.core.policy import (Policy,
                               Absent,
                               Present,
                               XOR,
                               NAND)

from yaybu.core.argument import (
    FullPath,
    String,
    Integer,
    Octal,
    File,
    )

class Special(Resource):

    """ A special file, as created by mknod. """

    name = FullPath()
    """ The full path to the special file on disk. """

    owner = String(default="root")
    """ The unix user who should own this special file. """

    group = String(default="root")
    """ The unix group who should own this special file. """

    mode = Octal(default=0644)
    """ The octal representation of the permissions for this special file. """

    type = String(default="fifo")
    """ One of the following strings:

      block
        create a block (buffered) special file
      character
        create a character (unbuffered) special file
      fifo
        create a fifo

    It defaults to fifo
    """

    major = Integer()
    """ The major number for the special file. If the type of the special file
    is block or character, then this must be specified. """

    minor = Integer()
    """ The minor number for the special file. If the type of the special file
    is block or character, then this must be specified. """

class SpecialAppliedPolicy(Policy):

    """ Ensure a block or character special file exists """

    name = "apply"
    default = True
    signature = (Present("name"),
                 Present("owner"),
                 Present("group"),
                 Present("mode"),
                 Present("type"),
                 Present("major"),
                 Present("minor"),
                 )


class SpecialRemovedPolicy(Policy):

    """ If the special file specified exists, remove it.

    You should only specify the special file to remove, the other fields are
    not needed """

    name = "remove"
    default = False
    signature = (Present("name"),
                 Absent("owner"),
                 Absent("group"),
                 Absent("mode"),
                 Absent("type"),
                 Absent("major"),
                 Absent("minor"),
                 )




# Copyright 2011 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module provides support functions for people wanting to use Yaybu
without Yay. This functionality is EXPERIMENTAL.
"""

import sys, inspect
from yaybu.core import resource

__bundle = None

def get_bundle():
    global __bundle
    if not __bundle:
        __bundle = resource.ResourceBundle()
    return __bundle

def reset_bundle():
    global __bundle
    __bundle = None

def create_wrapper(resource_type):
    """
    This creates a function for a Resource that performs string substitution
    and registers the  resource with a bundle.
    """
    def create_resource(**kwargs):
        caller = inspect.currentframe().f_back

        # Mutate kwargs based on local variables
        for k, v in kwargs.items():
            if isinstance(v, basestring):
                kwargs[k] = v.format(**caller.f_locals)

        # Create a Resource and add it to the bundle
        get_bundle().create(resource_type, kwargs)

    return create_resource

def register_resources():
    """
    This iterates over all known Resource Types and creates a proxy
    for them. The proxy will automatically format any strings it is passed
    with all variables in the local context. The proxy automatically
    registers the Resource with a resource bundle.

    It is called automatically when this module is imported.
    """
    module = sys.modules[__name__]
    for resource_type in resource.ResourceType.resources.keys():
        setattr(module, resource_type, create_wrapper(resource_type))

register_resources()




# Copyright 2011 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

# merci, twisted
def sibpath(path, sibling):
    """Return the path to a sibling of a file in the filesystem.

    This is useful in conjunction with the special __file__ attribute
    that Python provides for modules, so modules can load associated
    resource files.
    """
    return os.path.join(os.path.dirname(os.path.abspath(path)), sibling)




# Copyright 2011 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os, logging, json

from yaybu.core.protocol.server import HttpResource


class ChangeLogResource(HttpResource):

    leaf = True

    def render_POST(self, context, request, restpath):
        body = json.loads(request.rfile.read(int(request.headers["content-length"])))

        # Python logging seems to screw up if this isnt a tuple
        body['args'] = tuple(body.get('args', []))

        logrecord = logging.makeLogRecord(body)
        context.changelog.handle(logrecord)

        request.send_response(200, "OK")
        request.send_header("Content-Type", "application/octect-stream")
        request.send_header("Content-Length", "0")
        request.send_header("Content", "keep-alive")
        request.end_headers()




import os
import sys
from yaybu.harness import FakeChrootTestCase
from yaybu.core import error


class TestWatched(FakeChrootTestCase):

    def test_execute(self):
        self.fixture.check_apply("""
            hello.secret: world

            resources:
                - Execute:
                    name: test_watched
                    command: /bin/touch ${hello}
                    creates: /world
            """)

        self.failUnlessExists("/world")




# Copyright 2011 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import unittest

from yaybu.core import (resource,
                        policy,
                        argument,
                        provider,
                        error)

class R(resource.Resource):
    foo = argument.String()
    bar = argument.String()

class P1(policy.Policy):
    default = True
    name = 'p1'
    resource = R
    signature = [policy.Present("foo")]

class P2(policy.Policy):
    name = 'p2'
    resource = R
    signature = [policy.Present("bar")]

class P3(policy.Policy):
    name = 'p3'
    resource = R
    signature = [policy.Present("foo"),
                 policy.Present("bar")]

class Prov1(provider.Provider):
    policies = [P1,P2]

class Prov2(provider.Provider):
    policies = [P3]

class TestOrchestration(unittest.TestCase):

    def test_validate(self):
        r = R(name="1", foo="a", bar="b")
        pol = r.get_default_policy()
        self.assertEqual(r.get_default_policy().get_provider({}), Prov1)
        r = R(name="2")
        self.assertRaises(error.NonConformingPolicy, r.validate)
        r = R(name="3", policy="p1")
        self.assertRaises(error.NonConformingPolicy, r.validate)



# Copyright 2011 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import logging
import pickle
import subprocess
import StringIO

import yay
from yay.openers import Openers
from yay.errors import NotFound

from yaybu.core import resource
from yaybu.core.error import ParseError, MissingAsset, Incompatible
from yaybu.core.protocol.client import HTTPConnection
from yaybu.core.shell import Shell
from yaybu.core import change

logger = logging.getLogger("runcontext")

class RunContext(object):

    simulate = False
    ypath = ()
    verbose = 0

    def __init__(self, configfile, opts=None):
        self.path = []
        self.ypath = []
        self.options = {}
        self._config = None
        self._bundle = None

        self.resume = opts.resume
        self.no_resume = opts.no_resume
        self.user = opts.user

        self.host = opts.host
        self.connect_user = None
        self.port = None

        if self.host:
            if "@" in self.host:
                self.connect_user, self.host = self.host.split("@", 1)
            if ":" in self.host:
                self.host, self.port = self.host.split(":", 1)

        if os.path.exists("/etc/yaybu"):
            self.options = yay.load_uri("/etc/yaybu")

        if opts is not None:
            logger.debug("Invoked with ypath: %r" % opts.ypath)
            logger.debug("Environment YAYBUPATH: %r" % os.environ.get("YAYBUPATH", ""))
            self.simulate = opts.simulate
            self.ypath = opts.ypath
            self.verbose = opts.verbose

        if "PATH" in os.environ:
            for term in os.environ["PATH"].split(":"):
                self.path.append(term)

        if "YAYBUPATH" in os.environ:
            for term in os.environ["YAYBUPATH"].split(":"):
                self.ypath.append(term)

        if not len(self.ypath):
            self.ypath.append(os.getcwd())

        self.configfile = configfile

        self.setup_shell(opts.env_passthrough)
        self.setup_changelog()

    def setup_shell(self, environment):
        self.shell = Shell(context=self,
            verbose=self.verbose,
            simulate=self.simulate,
            environment=environment)

    def setup_changelog(self):
        self.changelog = change.ChangeLog(self)

    def locate(self, paths, filename):
        if filename.startswith("/"):
            return filename
        for prefix in paths:
            candidate = os.path.realpath(os.path.join(prefix, filename))
            logger.debug("Testing for existence of %r" % (candidate,))
            if os.path.exists(candidate):
                return candidate
            logger.debug("%r does not exist" % candidate)
        raise MissingAsset("Cannot locate file %r" % filename)

    def locate_file(self, filename):
        """ Locates a file by referring to the defined yaybu path. If the
        filename starts with a / then it is absolutely rooted in the
        filesystem and will be returned unmolested. """
        return self.locate(self.ypath, filename)

    def locate_bin(self, filename):
        """ Locates a binary by referring to the defined yaybu path and PATH. If the
        filename starts with a / then it is absolutely rooted in the
        filesystem and will be returned unmolested. """
        return self.locate(self.ypath + self.path, filename)

    def set_config(self, config):
        """ Rather than have yaybu load a config you can provide one """
        self._config = config

    def get_config(self):
        if self._config:
            return self._config

        try:
            c = yay.config.Config(searchpath=self.ypath)
            
            if self.host:
                extra = {
                    "yaybu": {
                        "host": self.host,
                        }
                    }
                c.add(extra)

            c.load_uri(self.configfile)

            self._config = c
            return c

        except yay.errors.Error, e:
            raise ParseError(e.get_string())

    def set_bundle(self, bundle):
        self._bundle = bundle

    def get_bundle(self):
        if self._bundle:
            return self._bundle

        cfg = self.get_config().mapping
        bundle = resource.ResourceBundle.create_from_yay_expression(cfg.get("resources").expand())
        bundle.bind()
        self._bundle = bundle

        return bundle

    def get_decrypted_file(self, filename):
        p = subprocess.Popen(["gpg", "-d", self.locate_file(filename)], stdout=subprocess.PIPE)
        return p.stdout

    def get_file(self, filename):
        try:
            return Openers(searchpath=self.ypath).open(filename)
        except NotFound, e:
            raise MissingAsset(str(e))


class RemoteRunContext(RunContext):

    def __init__(self, configfile, opts=None):
        self.connection = HTTPConnection()
        self.check_versions()
        super(RemoteRunContext, self).__init__(configfile, opts)

    def check_versions(self):
        self.connection.request("GET", "/about")
        rsp = self.connection.getresponse()

        if rsp.status != 200:
            self.features = []
            self.versions = {"Yaybu": "0", "yay": "0"}
        else:
            self.features = rsp.getheader("features", "").split(",")
            self.versions = {
                "Yaybu": rsp.getheader("Yaybu", "0"),
                "yay": rsp.getheader("yay", "0"),
                }

        import pkg_resources
        if pkg_resources.parse_version(self.versions["Yaybu"]) <= pkg_resources.parse_version("0"):
            raise Incompatible("You require a newer version of 'Yaybu' to deploy to this server")

        if pkg_resources.parse_version(self.versions["yay"]) <= pkg_resources.parse_version("0"):
            raise Incompatible("You require a newer version of 'yay' to deploy to this server")

    def setup_changelog(self):
        self.changelog = change.RemoteChangeLog(self)

    def get_config(self):
        self.connection.request("GET", "/config")
        rsp = self.connection.getresponse()
        c = yay.config.Config()
        c.add(pickle.loads(rsp.read()))
        return c

    def get_decrypted_file(self, filename):
        self.connection.request("GET", "/encrypted/" + filename)
        rsp = self.connection.getresponse()
        return rsp

    def get_file(self, filename):
        if filename.startswith("/"):
            return super(RemoteRunContext, self).get_file(filename)

        self.connection.request("GET", "/files/?path=" + filename)
        rsp = self.connection.getresponse()

        if rsp.status == 404:
            raise MissingAsset("Cannot fetch %r" % filename)

        return rsp




# Copyright 2011 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import unittest
from yay import String
from yaybu.providers.filesystem import files

class TestSecretTemplateArgs(unittest.TestCase):

    def setUp(self):
        self.ps = String()
        self.ps.add_secret('hello')

        class DummyResource:
            template_args = {}

        self.args = DummyResource.template_args
        self.provider = files.File(DummyResource)

    def test_no_secret(self):
        self.args['level1'] = dict(foo=1, bar=[1,2,3], baz=dict())
        self.failUnless(not self.provider.has_protected_strings())

    def test_secret_in_dict(self):
        self.args['level1'] = dict(foo=self.ps, bar=1)
        self.failUnless(self.provider.has_protected_strings())

    def test_secret_in_list(self):
        self.args['level1'] = [self.ps, 1, 2]
        self.failUnless(self.provider.has_protected_strings())

    def test_secret_in_dict_in_list(self):
        self.args['level1'] = [dict(foo=self.ps, bar=1)]
        self.failUnless(self.provider.has_protected_strings())




import os, shutil, stat

from yaybu.harness import FakeChrootTestCase
from yaybu.util import sibpath
from yaybu.core import error

test_execute_on_path = """
#!/bin/sh
touch /etc/test_execute_on_path
""".strip()

test_touches = """
#!/bin/sh
touch /etc/test_execute_touches
""".strip()


class TestExecute(FakeChrootTestCase):

    def test_execute_on_path(self):
        with self.fixture.open("/usr/bin/test_execute_on_path.sh", "w") as fp:
            fp.write(test_execute_on_path)
        self.fixture.chmod("/usr/bin/test_execute_on_path.sh", 0755)

        self.fixture.check_apply("""
            resources:
                - Execute:
                    name: test
                    command: test_execute_on_path.sh
                    creates: /etc/test_execute_on_path
            """)

    def test_execute_touches(self):
        """ test that command works as expected """
        with self.fixture.open("/usr/bin/test_touches.sh", "w") as fp:
            fp.write(test_touches)
        self.fixture.chmod("/usr/bin/test_touches.sh", 0755)

        self.fixture.check_apply("""
            resources:
                - Execute:
                    name: test
                    command: test_touches.sh
                    creates: /etc/test_execute_touches

            """)

    def test_command(self):
        """ test that commands works as expected """
        self.fixture.check_apply("""
            resources:
                - Execute:
                    name: test
                    command: touch /etc/foo
                    creates: /etc/foo
            """)
        self.failUnlessExists("/etc/foo")

    def test_commands(self):
        self.fixture.check_apply("""
            resources:
                - Execute:
                    name: test
                    commands:
                        - touch /etc/foo
                        - touch /etc/bar
                    creates: /etc/bar
            """)
        self.failUnlessExists("/etc/foo")
        self.failUnlessExists("/etc/bar")

    def test_cwd(self):
        """ test that cwd works as expected. """
        self.fixture.check_apply("""
            resources:
                - Execute:
                    name: test
                    command: touch foo
                    cwd: /etc
                    creates: /etc/foo
            """)
        self.failUnlessExists("/etc/foo")

    def test_environment(self):
        """ test that the environment is passed as expected. """
        self.fixture.check_apply("""
            resources:
                - Execute:
                    name: test
                    command: sh -c "touch $FOO"
                    environment:
                        FOO: /etc/foo
                    creates: /etc/foo
            """)
        self.failUnlessExists("/etc/foo")

    def test_returncode(self):
        """ test that the returncode is interpreted as expected. """
        self.fixture.check_apply("""
            resources:
                - Execute:
                    name: test-execute-returncode-true
                    command: /bin/true
                    touch: /test_returncode_marker_true
                - Execute:
                    name: test-execute-returncode-false
                    command: /bin/false
                    touch: /test_returncode_marker_false
                    returncode: 1
            """)

    def test_user(self):
        """ test that the user has been correctly set. """
        self.fixture.check_apply("""
            resources:
                - Execute:
                    name: test_user_change
                    command: python -c "import os; open('/foo','w').write(str(os.getuid())+'\\n'+str(os.geteuid()))"
                    user: nobody
                    creates: /foo
            """)

        with self.fixture.open("/foo") as fp:
            check_file = fp.read().split()

        self.failUnlessEqual(["65534"] * 2, check_file)

    def test_group(self):
        """ test that the group has been correctly set. """
        self.fixture.check_apply("""
            resources:
                - Execute:
                    name: test_group_change
                    command: python -c "import os; open('/foo','w').write(str(os.getgid())+'\\n'+str(os.getegid()))"
                    group: nogroup
                    creates: /foo
        """)

        with self.fixture.open("/foo") as fp:
            check_file = fp.read().split()
        self.failUnlessEqual(["65534"] * 2, check_file)

    def test_user_and_group(self):
        """ test that both user and group can be set together. """
        self.fixture.check_apply("""
            resources:
                - Execute:
                    name: test_group_change
                    command: python -c "import os; open('/foo','w').write('\\n'.join(str(x) for x in (os.getuid(),os.geteuid(),os.getgid(),os.getegid())))"
                    user: nobody
                    group: nogroup
                    creates: /foo
        """)

        with self.fixture.open("/foo") as fp:
            check_file = fp.read().split()
        self.failUnlessEqual(["65534"] * 4, check_file)

    def test_creates(self):
        """ test that the execute will not happen if the creates parameter
        specifies an existing file. """
        self.fixture.touch("/existing-file")
        self.fixture.check_apply("""
            resources:
              - Execute:
                  name: test_creates
                  command: touch /existing-file
                  creates: /existing-file
            """, expect=error.NothingChanged.returncode)

    def test_touch(self):
        """ test that touch does touch a file. """
        self.fixture.check_apply("""
            resources:
             - Execute:
                 name: test_touch
                 command: whoami
                 touch: /touched-file
            """)
        self.failUnlessExists("/touched-file")

    def test_touch_present(self):
        """ test that we do not execute if the touched file exists. """
        self.fixture.touch("/touched-file")
        self.fixture.check_apply("""
            resources:
             - Execute:
                 name: test_touch_present
                 command: touch /checkpoint
                 touch: /touched-file
            """, expect=254)

        self.failIfExists("/checkpoint")

    def test_touch_not_present(self):
        """ test that we do execute if the touched file does not exist. """
        self.fixture.check_apply("""
            resources:
             - Execute:
                 name: test_touch_not_present
                 command: touch /checkpoint
                 touch: /touched-file
            """)

        self.failUnlessExists("/checkpoint")
        self.failUnlessExists("/touched-file")

    def test_unless_true(self):
        """ test that an Execute wont execute if the unless expression
        is true """

        rv = self.fixture.apply("""
            resources:
              - Execute:
                  name: test
                  command: touch /test_unless_true
                  unless: /bin/true
            """)

        self.failUnlessEqual(rv, 254)

    def test_unless_false(self):
        """ test that an Execute will execute when the unless expression
        is false """

        self.fixture.check_apply("""
            resources:
              - Execute:
                  name: test
                  command: touch /test_unless_false
                  unless: /bin/false
                  creates: /test_unless_false
            """)

        self.failUnlessExists("/test_unless_false")

    def test_umask_022(self):
        self.fixture.check_apply("""
            resources:
              - Execute:
                  name: touch
                  command: touch /test_umask_022
                  umask: 022
                  creates: /test_umask_022
            """)
        self.failUnlessExists("/test_umask_022")

        mode = stat.S_IMODE(self.fixture.stat("/test_umask_022").st_mode)
        self.failUnlessEqual(mode, 0644)

    def test_umask_002(self):
        self.fixture.check_apply("""
            resources:
              - Execute:
                  name: touch
                  command: touch /test_umask_002
                  umask: 002
                  creates: /test_umask_002
            """)
        self.failUnlessExists("/test_umask_002")

        mode = stat.S_IMODE(self.fixture.stat("/test_umask_002").st_mode)
        self.failUnlessEqual(mode, 0664)

    def test_missing_binary(self):
        rv = self.fixture.apply("""
            resources:
              - Execute:
                  name: test_missing_binary
                  command: this_binary_definitely_doesnt_exist
            """)

        self.failUnlessEqual(rv, error.BinaryMissing.returncode)

    def test_missing_binary_absolute(self):
        rv = self.fixture.apply("""
            resources:
              - Execute:
                  name: test_missing_binary_absolute
                  command: /this_binary_definitely_doesnt_exist
            """)

        self.failUnlessEqual(rv, error.BinaryMissing.returncode)




# Copyright 2011 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import subprocess
import pickle
import sys
import os

import yay

from yaybu.core.protocol.server import Server, HttpResource, StaticResource, AboutResource
from yaybu.core.protocol.file import FileResource, EncryptedResource
from yaybu.core.protocol.changelog import ChangeLogResource

from yaybu.core.runner import Runner
from yaybu.core.runcontext import RunContext
from yaybu.core import error


class RemoteRunner(Runner):

    user_known_hosts_file = "/dev/null"
    identity_file = None
    strict_host_key_checking = "ask"
    interactive = True

    def __init__(self):
        self.ssh_options = {}

    def load_host_keys(self, filename):
        self.user_known_hosts_file = filename

    def load_system_host_keys(self):
        self.user_known_hosts_file = os.path.expanduser("~/.ssh/known_hosts")

    def set_identity_file(self, identity_file):
        self.identity_file = identity_file

    def set_missing_host_key_policy(self, policy):
        self.strict_host_key_checking = policy

    def set_interactive(self, interactive):
        self.interactive = interactive

    def run(self, ctx):
        # Get a bundle straight away - allows us to validate the bundle before
        # sending it remotely.
        try:
            bundle = ctx.get_bundle()
        except error.Error as e:
            ctx.changelog.error(str(e))
            return e.returncode

        command = ["ssh", "-A"]
        command.extend(["-o", "UserKnownHostsFile %s" % self.user_known_hosts_file])
        command.extend(["-o", "StrictHostKeyChecking %s" % self.strict_host_key_checking])

        for k, v in self.ssh_options.items():
            command.extend(["-o", "%s %s" % (k, v)])

        if self.identity_file:
            command.extend(["-o", "IdentityFile %s" % self.identity_file])

        if not self.interactive:
            command.extend(["-o", "BatchMode yes"])

        if ctx.connect_user:
            command.extend(["-l", ctx.connect_user])

        if ctx.port:
            command.extend(["-p", ctx.port])

        command.append(ctx.host)

        command.extend(["yaybu", "--remote"])

        if ctx.user:
            command.extend(["--user", ctx.user])

        if ctx.simulate:
            command.append("-s")

        if ctx.verbose:
            command.extend(list("-v" for x in range(ctx.verbose)))

        if ctx.resume:
            command.append("--resume")

        if ctx.no_resume:
            command.append("--no-resume")

        command.append("-")

        try:
            p = subprocess.Popen(command, stdin=subprocess.PIPE, stdout=subprocess.PIPE)

            root = HttpResource()
            root.put_child("config", StaticResource(pickle.dumps(ctx.get_config().get())))
            root.put_child("files", FileResource())
            root.put_child("encrypted", EncryptedResource())
            root.put_child("changelog", ChangeLogResource())
            root.put_child("about", AboutResource())

            Server(ctx, root, p.stdout, p.stdin).serve_forever()
            p.wait()

            if p.returncode == 255:
                raise error.ConnectionError("Could not connect to '%s'" % ctx.host)

            return p.returncode

        except error.Error, e:
            print >>sys.stderr, "Error: %s" % str(e)

            if p.poll() is None:
                try:
                    p.kill()
                except OSError:
                    if p.poll() is None:
                        raise

            return e.returncode

        # An unknown error occured
        return 253





from yaybu.harness import FakeChrootTestCase

from yaybu.core import resource
from yaybu.core import argument
from yaybu.core import error

import mock

class Test_Random(FakeChrootTestCase):

    # needs more work, particularly the File argument

    def resource_args(self, resource):
        for k, v in resource.__dict__.items():
            if isinstance(v, argument.Argument):
                yield k, v.__class__

    def resource_test_valid(self, resource):
        d = {}
        for name, klass in self.resource_args(resource):
            d[name] = klass._generate_valid()
        r = resource(**d)
        shell = mock.Mock()
        ctx = mock.Mock()
        ctx.simulate = True
        ctx.shell = shell
        ctx.shell.execute = mock.Mock(return_value=(0, '', ''))
        config = mock.Mock()
        try:
            r.apply(ctx, config)
        except error.Error:
            pass

    def NOtest_random(self):
        for r in resource.ResourceType.resources.values():
            self.resource_test_valid(r)



# Copyright 2011 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os, logging
import re

from yaybu.core.provider import Provider
from yaybu.core.error import CheckoutError
from yaybu import resources

log = logging.getLogger("git")

class Git(Provider):

    policies = (resources.checkout.CheckoutSyncPolicy,)

    REMOTE_NAME = "origin"

    @classmethod
    def isvalid(self, policy, resource, yay):
        return resource.scm and resource.scm.lower() == "git"

    def git(self, context, action, *args, **kwargs):
        command = [
            "git",
            #"--git-dir=%s" % os.path.join(self.resource.name, ".git"),
            #"--work-tree=%s" % self.resource.name,
            "--no-pager",
            action,
        ]

        command.extend(list(args))

        if os.path.exists(self.resource.name):
            cwd = self.resource.name
        else:
            cwd = os.path.dirname(self.resource.name)

        return context.shell.execute(command, user=self.resource.user, exceptions=False, cwd=cwd, **kwargs)

    def action_clone(self, context):
        """Adds resource.repository as a remote, but unlike a
        typical clone, does not check it out

        """
        if not os.path.exists(self.resource.name):
            rv, out, err = context.shell.execute(
                ["/bin/mkdir", self.resource.name],
                user=self.resource.user,
                exceptions=False,
            )

            if not rv == 0:
                raise CheckoutError("Cannot create the repository directory")

            rv, out, err = self.git(context, "init", self.resource.name)
            if not rv == 0:
                raise CheckoutError("Cannot initialise local repository.")

            self.action_set_remote(context)
            return True
        else:
            return False

    def action_set_remote(self, context):
        git_parameters = [
            "remote", "add",
            self.REMOTE_NAME,
            self.resource.repository,
        ]

        rv, out, err = self.git(context, *git_parameters)

        if not rv == 0:
            raise CheckoutError("Could not set the remote repository.")

    def action_update_remote(self, context):
        # Determine if the remote repository has changed
        remote_re = re.compile(self.REMOTE_NAME + r"\t(.*) \(.*\)\n")
        rv, stdout, stderr = self.git(context, "remote", "-v", passthru=True)
        remote = remote_re.search(stdout)
        if remote:
            if not self.resource.repository == remote.group(1):
                log.info("The remote repository has changed.")
                self.git(context, "remote", "rm", self.REMOTE_NAME)
                self.action_set_remote(context)
                return True
        else:
            raise CheckoutError("Cannot determine repository remote.")

        return False

    def action_checkout(self, context):
        # Determine which SHA is currently checked out.
        if os.path.exists(os.path.join(self.resource.name, ".git")):
            rv, stdout, stderr = self.git(context, "rev-parse", "--verify", "HEAD", passthru=True)
            if not rv == 0:
                head_sha = '0' * 40
            else:
                head_sha = stdout[:40]
                log.info("Current HEAD sha: %s" % head_sha)
        else:
            head_sha = '0' * 40

        changed = True
        # Revision takes precedent over branch
        if self.resource.revision:
            newref = self.resource.revision
            if newref == head_sha:
                changed = False
        elif self.resource.branch:
            rv, stdout, stderr = self.git(context, "ls-remote",
                                        self.resource.repository, passthru=True)
            if not rv == 0:
                raise CheckoutError("Could not query the remote repository")
            r = re.compile('([0-9a-f]{40})\t(.*)\n')
            refs_to_shas = dict([(b,a) for (a,b) in r.findall(stdout)])

            as_tag = "refs/tags/%s" % self.resource.branch
            as_branch = "refs/heads/%s" % self.resource.branch

            if as_tag in refs_to_shas.keys():
                annotated_tag = as_tag + "^{}"
                if annotated_tag in refs_to_shas.keys():
                    as_tag = annotated_tag
                newref = self.resource.branch
                changed = head_sha != refs_to_shas.get(as_tag)
            elif as_branch in refs_to_shas.keys():
                newref = "remotes/%s/%s" % (
                    self.REMOTE_NAME,
                    self.resource.branch
                )
                changed = head_sha != refs_to_shas.get(as_branch)
        else:
            raise CheckoutError("You must specify either a revision or a branch")

        if changed:
            rv, stdout, stderr = self.git(context, "checkout", newref)
            if not rv == 0:
                raise CheckoutError("Could not check out '%s'" % newref)

        return changed

    def apply(self, context):
        log.info("Syncing %s" % self.resource)

        # If necessary, clone the repository
        if not os.path.exists(os.path.join(self.resource.name, ".git")):
            self.action_clone(context)
        else:
            self.action_update_remote(context)

        # Always update the REMOTE_NAME remote
        self.git(context, "fetch", self.REMOTE_NAME)

        return self.action_checkout(context)



# Copyright 2011 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

from yaybu.core import provider
from yaybu.core import error
from yaybu import resources


def is_installed(context, resource):
    # work out if the package is already installed
    command = ["dpkg-query", "-W", "-f='${Status}'", resource.name]
    returncode, stdout, stderr = context.shell.execute(command,
                                                           exceptions=False, passthru=True)

    # if the return code is 0, dpkg is aware of the package
    if returncode == 0 and "install ok installed" in stdout:
        return True

    # if the return code is anything but zero or one, we have a problem
    if returncode > 1:
        raise error.DpkgError("%s search failed with return code %s" % (resource, returncode))

    return False


class AptInstall(provider.Provider):

    policies = (resources.package.PackageInstallPolicy,)

    @classmethod
    def isvalid(self, policy, resource, yay):
        if resource.version is not None:
            return False
        return super(AptInstall, self).isvalid(policy, resource, yay)

    def apply(self, context):
        if is_installed(context, self.resource):
            return False

        env = {
            "DEBIAN_FRONTEND": "noninteractive",
            }

        # the search returned 1, package is not installed, continue and install it
        command = ["apt-get", "install", "-q", "-y", self.resource.name]
        returncode, stdout, stderr = context.shell.execute(command, exceptions=False, env=env)

        if returncode != 0:
            raise error.AptError("%s failed with return code %d" % (self.resource, returncode))

        return True


class AptUninstall(provider.Provider):

    policies = (resources.package.PackageUninstallPolicy,)

    @classmethod
    def isvalid(self, policy, resource, yay):
        return super(AptUninstall, self).isvalid(policy, resource, yay)

    def apply(self, context):
        if not is_installed(context, self.resource):
            return False

        env = {
            "DEBIAN_FRONTEND": "noninteractive",
            }

        command = ["apt-get", "remove", "-q", "-y"]
        if self.resource.purge:
            command.append("--purge")
        command.append(self.resource.name)

        returncode, stdout, stderr = context.shell.execute(command, exceptions=False, env=env)

        if returncode != 0:
            raise error.AptError("%s failed to uninstall with return code %d" % (self.resource, returncode))

        return True





# Copyright 2011 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import os
import stat
import pwd
import grp
import logging

from yaybu import resources
from yaybu.core import provider, error

class Link(provider.Provider):

    policies = (resources.link.LinkAppliedPolicy,)

    @classmethod
    def isvalid(self, *args, **kwargs):
        # TODO: validation could provide warnings based on things
        # that are not the correct state at the point of invocation
        # but that will be modified by the yaybu script
        return super(Link, self).isvalid(*args, **kwargs)

    def _get_owner(self):
        """ Return the uid for the resource owner, or None if no owner is
        specified. """
        if self.resource.owner is not None:
            try:
                return pwd.getpwnam(self.resource.owner).pw_uid
            except KeyError:
                raise error.InvalidUser()

    def _get_group(self):
        """ Return the gid for the resource group, or None if no group is
        specified. """
        if self.resource.group is not None:
            try:
                return grp.getgrnam(self.resource.group).gr_gid
            except KeyError:
                raise error.InvalidGroup()

    def _stat(self):
        """ Extract stat information for the resource. """
        st = os.lstat(self.resource.name)
        uid = st.st_uid
        gid = st.st_gid
        mode = stat.S_IMODE(st.st_mode)
        return uid, gid, mode

    def apply(self, context):
        changed = False
        name = self.resource.name
        to = self.resource.to
        exists = False
        uid = None
        gid = None
        mode = None
        isalink = False

        if not os.path.exists(to):
            if not context.simulate:
                raise error.DanglingSymlink("Destination of symlink %r does not exist" % to)
            context.changelog.info("Destination of sylink %r does not exist" % to)

        owner = self._get_owner()
        group = self._get_group()

        try:
            linkto = os.readlink(name)
            isalink = True
        except OSError:
            isalink = False

        if not isalink or linkto != to:
            if os.path.lexists(name):
                context.shell.execute(["/bin/rm", "-rf", name])

            context.shell.execute(["/bin/ln", "-s", self.resource.to, name])
            changed = True

        try:
            linkto = os.readlink(name)
            isalink = True
        except OSError:
            isalink = False

        if not isalink and not context.simulate:
            raise error.OperationFailed("Did not create expected symbolic link")

        if isalink:
            uid, gid, mode = self._stat()

        if owner is not None and owner != uid:
            context.shell.execute(["/bin/chown", "-h", self.resource.owner, name])
            changed = True

        if group is not None and group != gid:
            context.shell.execute(["/bin/chgrp", "-h", self.resource.group, name])
            changed = True

        return changed

class RemoveLink(provider.Provider):

    policies = (resources.link.LinkRemovedPolicy,)

    @classmethod
    def isvalid(self, *args, **kwargs):
        return super(RemoveLink, self).isvalid(*args, **kwargs)

    def apply(self, context):
        if os.path.lexists(self.resource.name):
            if not os.path.islink(self.resource.name):
                raise error.InvalidProvider("%r: %s exists and is not a link" % (self, self.resource.name))
            context.shell.execute(["/bin/rm", self.resource.name])
            return True
        return False




import os
import sys
from yaybu.harness import FakeChrootTestCase
from yaybu.core import error

class TestArgumentParser(FakeChrootTestCase):

    def test_invalid_param(self):
        rv = self.fixture.apply("""
            resources:
                - Execute:
                    name: test_invalid_param
                    commandz: /bin/touch
            """)

        self.failUnlessEqual(rv, error.ParseError.returncode)

    def test_missing_arg(self):
        rv = self.fixture.apply("""
            resources:
                - Execute:
                    name: test_missing_arg
                    command: /bin/touch ${hello}
            """)

        self.failUnlessEqual(rv, error.ParseError.returncode)




# Copyright 2011 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import shlex

from yaybu.core import provider
from yaybu.core import error
from yaybu import resources


class Execute(provider.Provider):

    policies = (resources.execute.ExecutePolicy,)

    @classmethod
    def isvalid(self, *args, **kwargs):
        return super(Execute, self).isvalid(*args, **kwargs)

    def execute(self, shell, command, expected_returncode=None, passthru=False):
        # Filter out empty strings...
        cwd = self.resource.cwd or None
        env = self.resource.environment or None

        returncode, stdout, stderr = shell.execute(command, cwd=cwd, env=env, user=self.resource.user,
            group=self.resource.group, passthru=passthru, exceptions=False, umask=self.resource.umask)

        if not shell.simulate and expected_returncode != None and expected_returncode != returncode:
            raise error.CommandError("%s failed with return code %d" % (self.resource, returncode))

        return returncode

    def apply(self, context):
        if self.resource.creates is not None \
           and os.path.exists(self.resource.creates):
            #logging.info("%r: %s exists, not executing" % (self.resource, self.resource.creates))
            return False

        if self.resource.touch is not None \
                and os.path.exists(self.resource.touch):
            return False

        if self.resource.unless:
            if self.execute(context.shell, self.resource.unless, passthru=True) == 0:
                return False

        commands = [self.resource.command] if self.resource.command else self.resource.commands
        for command in commands:
            self.execute(context.shell, command, self.resource.returncode)

        if self.resource.touch is not None:
            context.shell.execute(["touch", self.resource.touch])

        return True





# Copyright 2011 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import shutil
from abc import ABCMeta, abstractmethod
from BaseHTTPServer import BaseHTTPRequestHandler
import StringIO

class RequestHandler(BaseHTTPRequestHandler):

    def __init__(self, rfile, wfile):
        self.rfile = rfile
        self.wfile = wfile

    def address_string(self):
        return 'stdio'

    def log_message(self, format, *args):
        # Uncomment to get HTTP request logs
        # super(RequestHandler, self).log_message(format, *args)
        return

    def handle_one_request(self):
        self.raw_requestline = self.rfile.readline()
        if not self.raw_requestline:
            self.close_connection = 1
            return False
        if not self.parse_request():
            return False

        path = self.path
        if "?" in path:
            path, self.getargs = path.split("?", 1)
        if "#" in path:
            path, self.bookmark = path.split("#", 1)
        self.path = filter(None, path.split("/"))

        return True

    def write_fileobj(self, fileobj):
        shutil.copyfileobj(fileobj, self.wfile)

    def send_error(self, code, message=None):
        """Send and log an error reply.

        We override the standard python code *soley* to keep the connection alive
        """

        try:
            short, long = self.responses[code]
        except KeyError:
            short, long = '???', '???'
        if message is None:
            message = short

        self.send_response(code, message)
        self.send_header("Content-Type", self.error_content_type)
        self.send_header('Connection', 'keepalive')
        self.end_headers()


class Server(object):

    def __init__(self, context, root, rfile=sys.stdin, wfile=sys.stdout):
        self.context = context
        self.root = root
        self.rfile = rfile
        self.wfile = wfile
        self.handlers = []

    def handle_request(self):
        # This will use BaseHTTPRequestHandler to parse HTTP headers off stdin,
        #   stdin is then ready to read any payload?
        r = RequestHandler(self.rfile, self.wfile)
        if not r.handle_one_request():
            return False

        node = self.root

        try:
            if r.path:
                segment, rest = r.path[0], r.path[1:]
                while segment:
                    node = node.get_child(segment)
                    if node.leaf:
                        break
                    if not rest:
                        break
                    segment, rest = rest[0], rest[1:]
 
            node.render(self.context, r, "/".join(rest))
        except Error, e:
            e.render(r, None)

        return True

    def serve_forever(self):
        while self.handle_request():
            pass


class Error(Exception):

    def render(self, request, post):
        request.send_error(self.error_code, self.error_string)


class NotFound(Error):

    error_code = 404
    error_string = "Resource not found"

class MethodNotSupported(Error):

    error_code = 501
    error_string = "Method not supported"


class HttpResource(object):

    leaf = False

    def __init__(self):
        self.children = {}

    def put_child(self, key, child):
        self.children[key] = child

    def get_child(self, key):
        if key in self.children:
            return self.children[key]
        raise NotFound(key)

    def render(self, yaybu, request, postpath):
        if not hasattr(self, "render_" + request.command):
            raise MethodNotSupported(request.command)
        getattr(self, "render_" + request.command)(yaybu, request, postpath)


class StaticResource(HttpResource):

    leaf = True

    def __init__(self, content):
        super(StaticResource, self).__init__()
        self.content = content

    def render_GET(self, yaybu, request, postpath):
        request.send_response(200, "OK")
        request.send_header("Content-Type", "application/json")
        request.send_header("Content-Length", str(len(self.content)))
        request.send_header("Content", "keep-alive")
        request.end_headers()
        request.write_fileobj(StringIO.StringIO(self.content))


class AboutResource(HttpResource):

    leaf = True

    def get_version(self, thing):
        """ Returns the version of a python egg. Returns 0 if it can't be determined """
        import pkg_resources
        try:
            return pkg_resources.get_distribution(thing).version
        except pkg_resources.DistributionNotFound:
            return "0"

    def render_GET(self, yaybu, request, postpath):    
        request.send_response(200, "OK")
        request.send_header("Content-Type", "text/plain")
        request.send_header("Content-Length", "0")
        request.send_header("Content", "keep-alive")
        request.send_header("Yaybu", self.get_version("Yaybu"))
        request.send_header("yay", self.get_version("yay"))
        request.end_headers()
        request.write_fileobj(StringIO.StringIO(""))




# Copyright 2011 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os, signal, shlex, subprocess, tempfile, time, shutil, StringIO
import testtools
from yaybu.core import error
from yaybu.util import sibpath

from yaybu.harness.fixture import Fixture

# Setup environment passthrough for chroot environment
# And turn on auditlog
yaybu_cfg = """
env-passthrough:
 - COWDANCER_ILISTFILE
 - FAKECHROOT
 - FAKECHROOT_VERSION
 - FAKECHROOT_BASE
 - FAKED_MODE
 - FAKEROOTKEY
 - LD_PRELOAD
 - LD_LIBRARY_PATH
auditlog:
   mode: file
"""

# A little SSH wrapper for faking SSH into a fakechroot
# (Obviously won't let us fake paramiko...)
sshwrapper = """
#! /usr/bin/env python
import os, sys
args = sys.argv[1:]
while args and not args[0] == "yaybu":
    del args[0]
if args:
    os.execvp(args[0], args)
""".strip()


class FakeChrootFixture(Fixture):

    """
    I provide a very simple COW userspace environment in which to test configuration

    I am used for some of Yaybu's internal tests.
    """

    firstrun = True

    fakerootkey = None

    testbase = os.getenv("YAYBU_TESTS_BASE", "base-image")
    test_network = os.environ.get("TEST_NETWORK", "0") == "1"

    def setUp(self):
        if self.firstrun:
            if not os.path.exists(self.testbase):
                self.build_environment()
            self.refresh_environment()

            # We only refresh the base environment once, so
            # set this on the class to make sure any other fixtures pick it up
            FakeChrootFixture.firstrun = False

        self.clone()

    def clone(self):
        self.chroot_path = os.path.realpath("tmp")
        subprocess.check_call(["cp", "-al", self.testbase, self.chroot_path])

        with self.open("/usr/bin/ssh", "w") as fp:
            fp.write(sshwrapper)
        self.chmod("/usr/bin/ssh", 0755)

        with self.open("/etc/yaybu", "w") as fp:
            fp.write(yaybu_cfg)
        self.chmod("/etc/yaybu", 0644)

    def cleanUp(self):
        self.cleanup_session()
        subprocess.check_call(["rm", "-rf", self.chroot_path])

    def reset(self):
        self.cleanUp()
        self.clone()

    def default_distro(self):
        options = {
            "Ubuntu 9.10": "karmic",
            "Ubuntu 10.04": "lucid",
            "Ubuntu 10.10": "maverick",
            "Ubuntu 11.04": "natty",
           }
        sundayname = open("/etc/issue.net","r").read().strip()
        return options[sundayname[:12]]

    def run_commands(self, commands, distro=None):
        for command in commands:
            command = command % dict(base_image=self.testbase, distro=distro)
            p = subprocess.Popen(shlex.split(command))
            if p.wait():
                raise SystemExit("Command failed")

    def build_environment(self):
        distro = self.default_distro()
        commands = [
            "fakeroot fakechroot -s debootstrap --variant=fakechroot --include=git-core,python-setuptools,python-dateutil,python-magic,ubuntu-keyring,gpgv %(distro)s %(base_image)s",
            "fakeroot fakechroot -s /usr/sbin/chroot %(base_image)s apt-get update",
            ]
        if not os.path.exists(self.testbase):
             self.run_commands(commands, distro)

    def refresh_environment(self):
        commands = [
             "rm -rf /usr/local/lib/python2.6/dist-packages/Yaybu*",
             "python setup.py sdist --dist-dir %(base_image)s",
             "fakeroot fakechroot -s /usr/sbin/chroot %(base_image)s sh -c 'easy_install /Yaybu-*.tar.gz'",
             ]
        self.run_commands(commands)

    def cleanup_session(self):
        if self.faked:
            os.kill(int(self.faked.strip()), signal.SIGTERM)
            self.faked = None

    def get_session(self):
        if self.fakerootkey:
            return self.fakerootkey

        p = subprocess.Popen(['faked-sysv'], stdout=subprocess.PIPE)

        stdout, stderr = p.communicate()
        self.fakerootkey, self.faked = stdout.split(":")
        return self.fakerootkey

    def write_temporary_file(self, contents):
        f = tempfile.NamedTemporaryFile(dir=os.path.join(self.chroot_path, 'tmp'), delete=False)
        f.write(contents)
        f.close()
        return f.name

    def call(self, command):
        env = os.environ.copy()
        env['FAKEROOTKEY'] = self.get_session()
        env['LD_PRELOAD'] = "/usr/lib/libfakeroot/libfakeroot-sysv.so"
        env['HOME'] = '/root/'

        # Meh, we inherit the invoking users environment - LAME.
        env['HOME'] = '/root'
        env['PWD'] = '/'
        env['LOGNAME'] = 'root'
        env['USERNAME'] = 'root'
        env['USER'] = 'root'

        chroot = ["fakechroot", "-s", "cow-shell", "/usr/sbin/chroot", self.chroot_path]
        retval = subprocess.call(chroot + command, cwd=self.chroot_path, env=env)

        self.wait_for_cowdancer()
        return retval

    def yaybu(self, *args):
        filespath = os.path.join(self.chroot_path, "tmp", "files")
        args = list(args)
        if self.test_network:
            args.insert(0, "localhost")
            args.insert(0, "--host")

        return self.call(["yaybu", "-d", "--ypath", filespath] + list(args))

    def simulate(self, *args):
        """ Run yaybu in simulate mode """
        args = ["--simulate"] + list(args)
        return self.yaybu(*args)

    def apply(self, contents, *args):
        path = self.write_temporary_file(contents)
        return self.yaybu(path, *args)

    def apply_simulate(self, contents):
        path = self.write_temporary_file(contents)
        return self.simulate(path)

    def check_apply(self, contents, *args, **kwargs):
        expect = kwargs.get('expect', 0)

        # Apply the change in simulate mode
        sim_args = list(args) + ["-s"]
        rv = self.apply(contents, *sim_args)
        if rv != expect:
            raise subprocess.CalledProcessError(rv, "Simulation failed: got rv %s" % rv)

        # Apply the change for real
        rv = self.apply(contents, *args)
        if rv != expect:
            raise subprocess.CalledProcessError(rv, "Apply failed: got rv %s" % rv)

        # If 'expect' isnt 0 then theres no point doing a no-changes check
        if expect != 0:
            return

        # If we apply the change again nothing should be changed
        rv = self.apply(contents, *args)
        if rv != error.NothingChanged.returncode:
            raise subprocess.CalledProcessError(rv, "Change still outstanding")

    def check_apply_simulate(self, contents):
        rv = self.apply_simulate(contents)
        if rv != 0:
            raise subprocess.CalledProcessError(rv, "Simulate failed rv %s" % rv)

    def wait_for_cowdancer(self):
        # give cowdancer a few seconds to exit (avoids a race where it delets another sessions .ilist)
        for i in range(20):
            if not os.path.exists(os.path.join(self.chroot_path, ".ilist")):
                break
            time.sleep(0.1)

    def exists(self, path):
        return os.path.exists(self._enpathinate(path))

    def isdir(self, path):
        return os.path.isdir(self._enpathinate(path))

    def mkdir(self, path):
        os.mkdir(self._enpathinate(path))

    def open(self, path, mode='r'):
        return open(self._enpathinate(path), mode)

    def touch(self, path):
        if not self.exists(path):
            with self.open(path, "w") as fp:
                fp.write("")

    def chmod(self, path, mode):
        self.call(["chmod", "%04o" % mode, self._enpathinate(path)])

    def readlink(self, path):
        relpath = os.path.relpath(os.readlink(self._enpathinate(path)), self.chroot_path)
        for x in (".", "/"):
            if relpath.startswith(x):
                relpath = relpath[1:]
        return "/" + relpath

    def symlink(self, source, dest):
        os.symlink(self._enpathinate(source), self._enpathinate(dest))

    def stat(self, path):
        return os.stat(self._enpathinate(path))

    def _enpathinate(self, path):
        return os.path.join(self.chroot_path, *path.split(os.path.sep))

    def get_user(self, user):
        users_list = open(self._enpathinate("/etc/passwd")).read().splitlines()
        users = dict(u.split(":", 1) for u in users_list)
        return users[user].split(":")

    def get_group(self, group):
        # Returns a tuple of group info if the group exists, or raises KeyError if it does not
        groups_list = open(self._enpathinate("/etc/group")).read().splitlines()
        groups = dict(g.split(":", 1) for g in groups_list)
        return groups[group].split(":")




from setuptools import setup, find_packages
import os

version = '0.1.19'

setup(name='Yaybu',
      version=version,
      url="http://yaybu.com/",
      description="Server deployment and configuration management in Python",
      long_description=open("README.rst").read(),
      author="Isotoma Limited",
      author_email="support@isotoma.com",
      license="Apache Software License",
      classifiers = [
          "Intended Audience :: System Administrators",
          "Operating System :: POSIX",
          "License :: OSI Approved :: Apache Software License",
      ],
      packages=find_packages(exclude=['ez_setup']),
      namespace_packages = ['yaybu'],
      include_package_data=True,
      zip_safe=False,
      install_requires=[
          'setuptools',
          'jinja2',
          'yay >= 0.0.41',
          'python-dateutil < 2.0',
      ],
      extras_require = {
          'test': ['testtools', 'discover', 'mock'],
          },
      entry_points = """
      [console_scripts]
      yaybu = yaybu.core.main:main
      [yaybu.resources]
      resources = yaybu.resources
      [yaybu.providers]
      providers = yaybu.providers
      """
      )



# Copyright 2011 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import pwd
try:
    import spwd
except ImportError:
    spwd = None
import grp

from yaybu.core import provider
from yaybu.core import error
from yaybu import resources

import logging

logger = logging.getLogger("provider")

class User(provider.Provider):

    policies = (resources.user.UserApplyPolicy,)

    @classmethod
    def isvalid(self, *args, **kwargs):
        return super(User, self).isvalid(*args, **kwargs)

    def get_user_info(self):
        fields = ("name", "passwd", "uid", "gid", "gecos", "dir", "shell")

        try:
            info_tuple = pwd.getpwnam(self.resource.name)
        except KeyError:
            info = dict((f, None) for f in fields)
            info["exists"] = False
            info['disabled-login'] = False
            info['disabled-password'] = False
            return info

        info = {"exists": True,
                "disabled-login": False,
                "disabled-password": False,
                }
        for i, field in enumerate(fields):
            info[field] = info_tuple[i]

        try:
            shadow = spwd.getspnam(self.resource.name)
            info['passwd'] = shadow.sp_pwd
            if shadow.sp_pwd == "!":
                info['disabled-login'] = True
        except KeyError:
            info['passwd'] = ''
            info['disabled-login'] = False

        return info

    def apply(self, context):
        if self.resource.password and self.resource.disabled_login:
            raise error.ParseError("Cannot specify password and disabled login for a user")

        info = self.get_user_info()

        if info['exists']:
            command = ['usermod']
            changed = False # we may not change anything yet
        else:
            command = ['useradd', '-N']
            changed = True # we definitely make a change

        if self.resource.fullname and info["gecos"] != self.resource.fullname:
            command.extend(["--comment", self.resource.fullname])
            changed = True

        if self.resource.password and not info["exists"]:
            command.extend(["--password", self.resource.password])
            changed = True

        if self.resource.home and info["dir"] != self.resource.home:
            command.extend(["--home", self.resource.home])
            changed = True

        if self.resource.uid and info["uid"] != self.resource.uid:
            command.extend(["--uid", str(self.resource.uid)])
            changed = True

        if self.resource.gid or self.resource.group:
            if self.resource.gid:
                gid = self.resource.gid
            else:
                try:
                    gid = grp.getgrnam(self.resource.group).gr_gid
                except KeyError:
                    if not context.simulate:
                        raise error.InvalidGroup("Group '%s' is not valid" % self.resource.group)
                    context.changelog.info("Group '%s' doesn't exist; assuming recipe already created it" % self.resource.group)
                    gid = "GID_CURRENTLY_UNASSIGNED"

            if gid != info["gid"]:
                command.extend(["--gid", str(gid)])
                changed = True

        if self.resource.groups:
            desired_groups = set(self.resource.groups)
            current_groups = set(g.gr_name for g in grp.getgrall() if self.resource.name in g.gr_mem)

            if self.resource.append and len(desired_groups - current_groups) > 0:
                if info["exists"]:
                    command.append("-a")
                command.extend(["-G", ",".join(desired_groups - current_groups)])
                changed = True
            elif not self.resource.append and desired_groups != current_groups:
                command.extend(["-G", ",".join(desired_groups)])
                changed = True

        if self.resource.shell != info["shell"]:
            command.extend(["--shell", str(self.resource.shell)])
            changed = True

        if self.resource.disabled_login and not info["disabled-login"]:
            command.extend(["--password", "!"])
            changed = True

        if info["exists"] is False and self.resource.system:
            command.extend(["--system"])
            changed = True

        command.extend(["-m", self.resource.name])

        if changed:
            returncode, stdout, stderr = context.shell.execute(command, exceptions=False)
            if returncode != 0:
                raise error.UserAddError("useradd returned error code %d" % returncode)
        return changed


class UserRemove(provider.Provider):

    policies = (resources.user.UserRemovePolicy,)

    @classmethod
    def isvalid(self, *args, **kwargs):
        return super(UserRemove, self).isvalid(*args, **kwargs)

    def apply(self, context):
        try:
            existing = pwd.getpwnam(self.resource.name.encode("utf-8"))
        except KeyError:
            # If we get a key errror then there is no such user. This is good.
            return False

        command = ["userdel", self.resource.name]

        returncode, stdout, stderr = context.shell.execute(command)
        if returncode != 0:
            raise error.UserAddError("Removing user %s failed with return code %d" % (self.resource, returncode))

        return True





# Copyright 2011 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import policy
import yaml

class EventState(object):

    """ Represents the current state of events """

    save_file = "events.saved"
    """ The file to save to.  This is touched by the runner. """


    overrides = {}
    """ A mapping of resource ids to the overridden policy name for that
    resource, if there is one. """

    def __init__(self, load=False):
        self.loaded = not load
        self.overrides = {}
        self.simulate = False

    def load(self):
        if self.loaded:
            return
        if os.path.exists(self.save_file):
            self.overrides = yaml.load(open(self.save_file))
        self.loaded = True

    def override(self, resource, policy):
        self.load()
        self.overrides[resource.id] = policy
        self.save()

    def clear_override(self, resource):
        self.load()
        if resource.id in self.overrides:
            del self.overrides[resource.id]
            self.save()

    def overridden_policy(self, resource):
        """ Return the policy class for this resource, or None if there is not
        an overridden policy. """
        if resource.id in self.overrides:
            policy_name = self.overrides[resource.id]
            return resource.policies[policy_name]
        else:
            return None

    def policy(self, resource):
        self.load()
        selected = self.overridden_policy(resource)
        if selected is None:
            if resource.policy is not None:
                selected = resource.policy.literal_policy(resource)
            else:
                selected = resource.policies.default()
        return selected(resource)

    def save(self):
        if not self.simulate:
            yaml.dump(self.overrides, open(self.save_file, "w"),  default_flow_style=False)


# module level global to preserve event state
# yes this is ugly
# alternatives may be uglier
state = EventState()

def reset():
    global state
    state = EventState()



# Copyright 2011 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pdb, socket, sys


class Rdb(pdb.Pdb):

    def __init__(self, port=4444):
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

        self.socket.bind(("127.0.0.1", port))
        self.socket.listen(1)

        client, address = self.socket.accept()
        handle = client.makefile('rw')
        pdb.Pdb.__init__(self, completekey='tab', stdin=handle, stdout=handle)

    def do_continue(self, arg):
        self.socket.shutdown(socket.SHUT_RDWR)
        self.socket.close()
        self.socket = None
        self.set_continue()
        return 1

    def __enter__(self):
        return self

    def __exit__(self, *args):
        if self.socket:
            self.socket.shutdown(socket.SHUT_RDWR)
            self.socket.close()
        return False

def set_trace():
    Rdb().set_trace(sys._getframe().f_back)

def post_mortem(exc_traceback=None):
    if exc_traceback is None:
       exc_type, exc_value, exc_traceback = sys.exc_info()

    with Rdb() as p:
        p.reset()
        p.interaction(None, exc_traceback)

def pm():
    post_mortem(sys.last_traceback)




# Copyright 2011 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import subprocess
import StringIO
import change
import error
import os, getpass, pwd, grp, select
import shlex

from yay import String

class Command(String):
    """ Horrible horrible cludge """
    pass


class Handle(object):

    def __init__(self, handle, callback=None):
        self.handle = handle
        self.callback = callback
        self._output = []

    def fileno(self):
        return self.handle.fileno()

    def read(self):
        data = os.read(self.fileno(), 1024)
        if data == "":
            self.handle.close()
            return False

        self._output.append(data)

        if self.callback:
            for l in data.splitlines():
                self.callback(l + "\r")

        return True

    def isready(self):
        return bool(self.handle)

    @property
    def output(self):
        out = ''.join(self._output)
        return out


class ShellCommand(change.Change):

    """ Execute and log a change """

    def __init__(self, command, shell, stdin, cwd=None, env=None, env_passthru=None, verbose=0, passthru=False, user=None, group=None, simulate=False, logas=None, umask=None):
        self.command = command
        self.shell = shell
        self.stdin = stdin
        self.cwd = cwd
        self.env = env
        self.env_passthru = env_passthru
        self.verbose = verbose
        self.passthru = passthru
        self.simulate = simulate
        self.logas = logas
        self._generated_env = {}

        self.user = None
        self.uid = None
        self.group = None
        self.gid = None
        self.homedir = None

        self.umask = umask

        if self.simulate and not self.passthru:
            # For now, we skip this setup in simulate mode - not sure it will ever be possible
            return

        self.user = user
        if user:
            u = pwd.getpwnam(user)
            self.uid = u.pw_uid
            self.homedir = u.pw_dir
        else:
            self.uid = None
            self.homedir = pwd.getpwuid(os.getuid()).pw_dir
            self.user = pwd.getpwuid(os.getuid()).pw_name

        self.group = group
        if group:
            self.gid = grp.getgrnam(self.group).gr_gid

    def preexec(self):
        if self.gid is not None:
            if self.gid != os.getgid():
                os.setgid(self.gid)
            if self.gid != os.getegid():
                os.setegid(self.gid)

        if self.uid is not None:
            if self.uid != os.getuid():
                os.setuid(self.uid)
            if self.uid != os.geteuid():
                os.seteuid(self.uid)

        if self.umask:
            os.umask(self.umask)

        os.environ.clear()
        os.environ.update(self._generated_env)

    def communicate(self, p, stdout_fn=None, stderr_fn=None):
        if p.stdin:
            p.stdin.flush()
            p.stdin.close()

        stdout = Handle(p.stdout, stdout_fn)
        stderr = Handle(p.stderr, stderr_fn)

        # Initial readlist is any handle that is valid
        readlist = [h for h in (stdout, stderr) if h.isready()]

        while readlist:
            try:
                rlist, wlist, xlist = select.select(readlist, [], [])
            except select.error, e:
                if e.args[0] == errno.EINTR:
                    continue
                raise

            # Read from all handles that select told us can be read from
            # If they return false then we are at the end of the stream
            # and stop reading from them
            for r in rlist:
                if not r.read():
                    readlist.remove(r)

        returncode = p.wait()

        return returncode, stdout.output, stderr.output

    def _tounicode(self, l):
        """ Ensure all elements of the list are unicode """
        def uni(x):
            if type(x) is type(u""):
                return x
            return unicode(x, "utf-8")
        return map(uni, l)

    def apply(self, renderer):
        if isinstance(self.command, Command):
            logas = self.command.as_list(secret=True)
            command = self.command.as_list(secret=False)
        elif isinstance(self.command, String):
            logas = shlex.split(self.command.protected.encode("UTF-8"))
            command = shlex.split(self.command.unprotected.encode("UTF-8"))
        elif isinstance(self.command, list):
            logas = command = self.command[:]
        elif isinstance(self.command, basestring):
            logas = command = shlex.split(self.command.encode("UTF-8"))

        command = self._tounicode(command)
        logas = self._tounicode(logas)

        renderer.passthru = self.passthru
        renderer.command(self.logas or logas)

        env = {
            "HOME": self.homedir,
            "LOGNAME": self.user,
            "PATH": "/usr/local/sbin:/usr/local/bin:/usr/sbin:/usr/bin:/sbin:/bin",
            "SHELL": "/bin/sh",
            }

        if self.env_passthru:
            for var in self.env_passthru:
                if var in os.environ:
                    env[var] = os.environ[var]

        if self.env:
            env.update(self.env)

        self._generated_env = env

        if self.simulate and not self.passthru:
            self.returncode = 0
            self.stdout = ""
            self.stderr = ""
            return

        command_exists = True
        if command[0].startswith("./"):
            if len(command[0]) <= 2:
                command_exists = False
            if not os.path.exists(os.path.join(self.cwd, command[0][2:])):
                command_exists = False
        elif command[0].startswith("/"):
            if not os.path.exists(command[0]):
                command_exists = False
        else:
            for path in env["PATH"].split(":"):
                if os.path.exists(os.path.join(path, command[0])):
                    break
            else:
                command_exists = False

        if not command_exists:
            if not self.simulate:
                raise error.BinaryMissing("Command '%s' not found" % command[0])
            self.context.changelog.info("Command '%s' not found; assuming this recipe will create it" % command[0])
 
        try:
            p = subprocess.Popen(command,
                                 shell=self.shell,
                                 stdout=subprocess.PIPE,
                                 stderr=subprocess.PIPE,
                                 cwd=self.cwd,
                                 env=None,
                                 preexec_fn=self.preexec,
                                 )
            self.returncode, self.stdout, self.stderr = self.communicate(p, renderer.stdout, renderer.stderr)
            renderer.output(p.returncode)
        except Exception, e:
            renderer.exception(e)
            raise

class ShellTextRenderer(change.TextRenderer):

    """ Render a ShellCommand on a textual changelog. """

    renderer_for = ShellCommand
    passthru = False

    def command(self, command):
        if not self.passthru:
            self.logger.notice(u"# " + u" ".join(command))

    def output(self, returncode):
        if self.verbose >= 1 and returncode != 0 and not self.passthru:
            self.logger.notice("returned %s", returncode)

    def stdout(self, data):
       if self.verbose >= 2 and not self.passthru:
            self.logger.info(data)

    def stderr(self, data):
        if self.verbose >= 1:
            self.logger.info(data)

    def exception(self, exception):
        self.logger.notice("Exception: %r" % exception)

class Shell(object):

    """ This object wraps a shell in yet another shell. When the shell is
    switched into "simulate" mode it can just print what would be done. """

    def __init__(self, context, verbose=0, simulate=False, environment=None):
        self.simulate = context.simulate
        self.verbose = context.verbose
        self.context = context

        self.environment = ["SSH_AUTH_SOCK"]
        if environment:
            self.environment.extend(environment)

    def locate_bin(self, filename):
        return self.context.locate_bin(filename)

    def execute(self, command, stdin=None, shell=False, passthru=False, cwd=None, env=None, exceptions=True, user=None, group=None, logas=None, umask=None):
        cmd = ShellCommand(command, shell, stdin, cwd, env, self.environment, self.verbose, passthru, user, group, self.simulate, logas, umask)
        self.context.changelog.apply(cmd)
        if exceptions and cmd.returncode != 0:
            self.context.changelog.info(cmd.stdout)
            self.context.changelog.notice(cmd.stderr)
            raise error.SystemError(cmd.returncode)
        return (cmd.returncode, cmd.stdout, cmd.stderr)




from yaybu.harness import FakeChrootTestCase
from yaybu.core.error import MissingDependency


class SubversionMissingTest(FakeChrootTestCase):

    def test_missing_svn(self):
        rv = self.fixture.apply("""
           resources:
               - Checkout:
                   scm: subversion
                   name: /dest
                   repository: /source
           """)
        self.assertEqual(MissingDependency.returncode, rv)

